from youtube_autonomous.shortcodes.enums import ShortcodeType


class ShortcodeTag:
    """
    Class that represent a shortcode tag to implement with
    the ShortcodeParser. This is just to let the parser know
    if it is a block-scoped shortcode tag, a simple shortcode
    tag and some more information needed.
    """
    def __init__(self, tag: str, type: ShortcodeType):
        """
        Initializes a shortcode tag object. The 'tag' parameter represents
        the shortcode name [tag], and the 'type' parameter is to point if
        the shortcode includes some text inside it [tag] ... [/tag] or if
        it is a simple one [tag].
        """
        if not tag:
            raise Exception('No "tag" provided.')
        
        if not type:
            raise Exception('No "type" provided.')

        if not isinstance(tag, str):
            raise Exception('Provided "tag" is not a string.')
        
        if not isinstance(type, ShortcodeType):
            raise Exception('Provided "type" is not a ShortcodeType.')
        
        self.tag = tag
        self.type = type

    def is_block_scoped(self):
        """
        Returns True if the shortcode is a block scoped one, that
        should look like this: [tag] ... [/tag], or False if not.
        """
        return self.type == ShortcodeType.BLOCK




from enum import Enum


class YoutubeChannelId(Enum):
    """
    Our specific channel ids.
    """
    STOCK = 'UCS49NBDOLkrahf6UgYiL4SQ'
    """
    Stock videos to be used as main part of the video background.
    """
    MEMES = 'UCiz62GOuKsIw3n7i3Cs4pRA'
    """
    Meme videos to enhance the experience by overlaying them or
    putting inline.
    """
    SOUNDS = 'UCF1vrXoRtVfUY05fCnmfNqw'
    """
    Sound videos to be downloaded as audio and used to enhance the
    video experience by using them in different specific situations.
    """
    MUSIC = 'UCDLfH57hn3K7lhmcAAqp9Mw'
    """
    Music videos to be downloaded as audio and used to enhance the
    video experience by using them as background music for the 
    whole video or for some specific moments.
    """


from youtube_autonomous.segments.enums import SegmentElementField, SegmentElementType, SegmentElementDuration, SegmentElementStart, SegmentElementMode, SegmentElementOrigin


class SegmentEnhancementValidator:
    """
    This class is to validate the Segment enhancement elements that
    are terms that the user register to enhance (to improve) the
    project video experience.

    These terms need to have a valid structure and that's what we
    check here.
    """
    def is_valid(enhancement_terms):
        """
        We will receive the content of a enhancement term and will raise an
        Exception if some structure element or value is not valid according
        to our rules.

        TODO: Write a little bit more about what we are receiving here.
        """
        # TODO: Maybe refactor the terms to simplify them
        for enhancement_term_key in enhancement_terms:
            enhancement_term = enhancement_terms[enhancement_term_key]
            for type in enhancement_term:
                content = enhancement_term[type]
                if type not in SegmentElementType.ALL.value:
                    # TODO: Make custom exception
                    raise Exception('Segment element type "' + type + '" not accepted.')

                # Lets check values
                active = content.get(SegmentElementField.ACTIVE.value)
                keywords = content.get(SegmentElementField.KEYWORDS.value)
                start = content.get(SegmentElementField.START.value)
                duration = content.get(SegmentElementField.DURATION.value)
                mode = content.get(SegmentElementField.MODE.value)
                origin = content.get(SegmentElementField.ORIGIN.value)

                if not 'active' in content:
                    # TODO: Make custom exception
                    raise Exception('No "' + SegmentElementField.ACTIVE.value + '" field provided in "' + enhancement_term_key + '" term. This field is mandatory and must be "true" if you want to be applied or "false" if not, but it must be set.')

                # Validate fields (even if it is not active)
                if not keywords:
                    # TODO: Make custom exception
                    raise Exception('No "' + SegmentElementField.KEYWORDS.value + '" provided in "' + enhancement_term_key + '" term. This field is mandatory.')

                if not start or start not in SegmentElementStart.ALL.value:
                    # TODO: Make custom exception
                    raise Exception('No "' + SegmentElementField.START.value + '" provided or not valid in "' + enhancement_term_key + '" term. These are the valid ones: "' + ', '.join(SegmentElementStart.ALL.value) + '"')

                
                if not (duration and (duration in [SegmentElementDuration.END_OF_CURRENT_WORD.value, SegmentElementDuration.FILE_DURATION.value] or duration.startswith(SegmentElementDuration.END_OF_SUBSEQUENT_WORD.value))):
                    # TODO: Make custom exception
                    raise Exception('No "' + SegmentElementField.DURATION.value + '" provided or not valid in "' + enhancement_term_key + '" term. These are the valid ones: "' + ', '.join(SegmentElementDuration.ALL.value) + '"')
                
                if duration.startswith(SegmentElementDuration.END_OF_SUBSEQUENT_WORD.value):
                    try:
                        aux = duration.split('_')
                        int(aux[len(aux) - 1])
                    except:
                        raise Exception('The "' + SegmentElementField.DURATION.value + '" field you provided must be like "' + SegmentElementDuration.END_OF_SUBSEQUENT_WORD.value + 'X" format, where the "X" must be an integer number that represent the subsequent word index that must be used in calculations.')

                if not mode or mode not in SegmentElementMode.ALL.value:
                    # TODO: Make custom exception
                    raise Exception('No "' + SegmentElementField.MODE.value + '" provided or not valid in "' + enhancement_term_key + '" term. These are the valid ones: "' + ', '.join(SegmentElementMode.ALL.value) + '"')

                if not origin or origin not in SegmentElementOrigin.ALL.value:
                    # TODO: Make custom exception
                    raise Exception('No "' + SegmentElementField.ORIGIN.value + '" provided or not valid in "' + enhancement_term_key + '" term. These are the valid ones: "' + ', '.join(SegmentElementOrigin.ALL.value) + '"')


from enum import Enum


class SegmentType(Enum):
    """
    These Enums represent the types that a Segment could be, allowing
    us to check and detect if it is valid.
    """
    # Interesting: https://docs.python.org/3/howto/enum.html
    # TODO: Remove 'MY_STOCK' when refactored (now use CUSTOM_STOCK)
    MY_STOCK = 'my_stock'
    CUSTOM_STOCK = 'custom_stock'
    """
    Stock videos but extracted from our own custom sources.
    """
    STOCK = 'stock'
    """
    Stock videos extracted from external stock platforms.
    """
    AI_IMAGE = 'ai_image'
    IMAGE = 'image'
    YOUTUBE_VIDEO = 'youtube_video'
    YOUTUBE_SEARCH = 'youtube_search'
    GOOGLE_SEARCH = 'google_search'
    TEXT = 'text'
    MEME = 'meme'

    @classmethod
    def get_premade_types(cls):
        """
        Returns a list containing all the Segment Types thar are 
        premades.
        """
        return [SegmentType.YOUTUBE_SEARCH, SegmentType.GOOGLE_SEARCH]

    @classmethod
    def is_valid(cls, type: str):
        """
        Returns True if the provided 'type' is a valid SegmentType enum
        or False if not.
        """
        if not type:
            raise Exception('No "type" provided.')
        
        if not isinstance(type, str):
            raise Exception('The "type" parameter provided is not a string.')
        
        try:
            SegmentType(type)
            return True
        except Exception:
            return False
    
class SegmentField(Enum):
    """
    These Enums represent the fields that a Segment has, allowing us
    to check that any required field is provided and/or to detect 
    which one is missing.

    Examples: TYPE, KEYWORDS, URL, etc.
    """
    # Interesting: https://docs.python.org/3/howto/enum.html
    TYPE = 'type'
    KEYWORDS = 'keywords'
    URL = 'url'
    TEXT = 'text'
    VOICE = 'voice'
    DURATION = 'duration'
    AUDIO_NARRATION_FILENAME = 'audio_narration_filename'
    
    @classmethod
    def get_all(cls):
        """
        Returns a list containing all the registered SegmentField enums.
        """
        return [SegmentField.TYPE, SegmentField.KEYWORDS, SegmentField.URL, SegmentField.TEXT, SegmentField.VOICE, SegmentField.DURATION, SegmentField.AUDIO_NARRATION_FILENAME]
    
class SegmentBuildingField(Enum):
    """
    The fields that are used when building the segment and are
    not provided by the user in the initial segment json data.
    """
    TRANSCRIPTION = 'transcription'
    AUDIO_FILENAME = 'audio_filename'
    AUDIO_CLIP = 'audio_clip'
    VIDEO_FILENAME = 'video_filename'
    VIDEO_CLIP = 'video_clip'
    FULL_FILENAME = 'full_filename'
    FULL_CLIP = 'full_clip'
    # TODO: What about 'status', 'music_filename' and any other (if existing) (?)

class ProjectField(Enum):
    """
    The fields that are used to handle the project information.
    """
    STATUS = 'status'
    """
    The current status of the project, that must be a ProjectStatus
    enum.
    """
    SEGMENTS = 'segments'
    """
    The array that contains all the information about each one of 
    his segments.
    """

class SegmentElementType(Enum):
    """
    These Enums represent the type of elements we can use to enhance
    a segment.
    """
    MEME = 'meme'
    STICKER = 'sticker'
    IMAGE = 'image'
    SOUND = 'sound'
    GREEN_SCREEN = 'green-scren'
    EFFECT = 'effect'
    ALL = [MEME, STICKER, IMAGE, SOUND, GREEN_SCREEN, EFFECT]
    """
    This enum must be used only for validation process. This has been made to 
    check if some provided type is valid as it is one of the valid and 
    available ones.
    """

class SegmentElementField(Enum):
    """
    These Enums represent the type of fields we can use with the
    segment enhancement elements to fit correctly the segment.
    """
    ACTIVE = 'active'
    KEYWORDS = 'keywords'
    DURATION = 'duration'
    START = 'start'
    MODE = 'mode'
    ORIGIN = 'origin'
    ALL = [ACTIVE, KEYWORDS, DURATION, START, MODE, ORIGIN]
    """
    This enum must be used only for validation process. This has been made to
    check if some provided enhancement field is valid or not and that all fields
    are given.
    """

class SegmentConstants(Enum):
    """
    These Enums are constants we need to use during the project
    and segments processing.
    """
    FILE_DURATION = 9999
    """
    This is the duration, as int value, that will be considered as
    the original video duration and will be used to set the original
    clip (that can be downloaded or load from local storage) duration
    as the final duration. This amount is just to be recognized and 
    will be replaced, when file processed, by the real duration.
    """

class SegmentElementMode(Enum):
    """
    These Enums represent the different ways in which the project
    segment elements can be built according to the way they are
    included in the segment.
    """
    INLINE = 'inline'
    """
    Those segment elements that will be displayed in 'inline' mode, that
    means they will interrupt the main video, be played, and then go back
    to the main video. This will modify the clip length, so we need to 
    refresh the other objects start times.
    """
    OVERLAY = 'overlay'
    """
    Those segment elements that will be displayed in 'overlay' mode, that
    means they will be shown in the foreground of the main clip, changing
    not the main video duration, so they don't force to do any refresh.
    """
    DEFAULT = INLINE
    """
    This is the default value I want (me, the developer, is the one who
    sets it by now). This one will be pointing to another of the available
    Enums. This Enum has been created to make it easy building objects in
    code and having the default value in only one strict place.
    """
    ALL = [INLINE, OVERLAY]
    """
    This is for validation purposes only.
    """

class SegmentElementOrigin(Enum): 
    """
    These Enums represent the way in which the project segment elements
    are searched and obtained from.
    """   
    YOUTUBE = 'youtube'
    """
    Those segment elements that will be searched in Youtube and downloaded
    if found and available.
    """
    FILE = 'file'
    """
    Those segment elements that will be load from local storage. The 
    location must be set in the segment building object.
    """
    DEFAULT = YOUTUBE
    """
    This is the default value I want (me, the developer, is the one who
    sets it by now). This one will be pointing to another of the available
    Enums. This Enum has been created to make it easy building objects in
    code and having the default value in only one strict place.
    """
    ALL = [YOUTUBE, FILE]
    """
    This is for validation purposes only.
    """

class SegmentElementStart(Enum):
    """
    These Enums represent the moment of the segment in which the 
    elements will be included.
    """
    START_OF_CURRENT_WORD = 'start_of_current_word'
    """
    This will make the segment object appear in the same moment that the 
    word that is before the shortcode starts (attending to transcription
    'start' timestamped field).
    """
    END_OF_CURRENT_WORD = 'end_of_current_word'
    """
    This will make the segment object appear in the same moment that the
    word that is before the shortcode ends (attending to transcription
    'end' timestamped field).
    """
    BEFORE_NEXT_WORD = 'before_next_word'
    """
    This will make the segment object appear in between the word that is
    before the shortcode and the next one. This one should start in the 
    silence gap between words.
    """
    DEFAULT = START_OF_CURRENT_WORD
    ALL = [START_OF_CURRENT_WORD, END_OF_CURRENT_WORD, BEFORE_NEXT_WORD]
    """
    This is for validation purposes only.
    """

class SegmentElementDuration(Enum):
    """
    These Enums represent the time, during a segment lifetime, that
    the elements are going to last (be shown).
    """
    END_OF_CURRENT_WORD = 'end_of_current_word'
    """
    This will make the segment object last until the end of the word
    that this literal says. This has a ':' because it must have
    a number after the ':' to work. The number means the amount of
    words that the segment object will last. If 'end_of_current_word:0', it 
    will last until the end of the word that is just before the 
    shortcode. If 'end_of_current_word:2', it will last until the end of the
    next of the next word (2nd word after the shortcode).
    """
    END_OF_SUBSEQUENT_WORD = 'end_of_subsequent_word_'
    """
    This will make the segment object last until the end of one of
    the subsequent words. This string will be end with a number 
    just after the last underscore, and that will be the number of
    words we need to wait to end. For example, if the value is
    'end_of_subsequent_word_2', we will count 2 more words to
    wait for the end of that one (if available).
    """
    FILE_DURATION = 'file_duration'
    """
    This will make the segment last the clip duration. It will be
    considered when the file is downloaded, and that duration will
    be flagged as 9999. This is for videos or audios that have a
    duration based on file.
    """
    ALL = [END_OF_CURRENT_WORD, END_OF_SUBSEQUENT_WORD, FILE_DURATION]
    """
    This is for validation purposes only.
    """

class SegmentStatus(Enum):
    """
    The current segment status defined by this string.
    """
    TO_START = 'to_start'
    IN_PROGRESS = 'in_progress'
    FINISHED = 'finished'

    @classmethod
    def is_valid(cls, status: str):
        """
        Returns True if the provided 'status' is a valid SegmentStatus enum
        or False if not.
        """
        if not status:
            raise Exception('No "status" provided.')
        
        if not isinstance(status, str):
            raise Exception('The "status" parameter provided is not a string.')
        
        try:
            SegmentStatus(status)
            return True
        except Exception:
            return False

class ProjectStatus(Enum):
    """
    The current project status defined by this string.
    """
    TO_START = 'to_start'
    IN_PROGRESS = 'in_progress'
    FINISHED = 'finished'

    @classmethod
    def is_valid(cls, status: str):
        """
        Returns True if the provided 'status' is a valid ProjectStatus enum
        or False if not.
        """
        if not status:
            raise Exception('No "status" provided.')
        
        if not isinstance(status, str):
            raise Exception('The "status" parameter provided is not a string.')
        
        try:
            ProjectStatus(status)
            return True
        except Exception:
            return False


from youtube_autonomous.database.database_handler import DatabaseHandler
from youtube_autonomous.segments.validation.segment_validator import SegmentValidator
from youtube_autonomous.project import Project
from youtube_autonomous.segments.builder.config import DEFAULT_SEGMENT_PARTS_FOLDER
from yta_general_utils.temp import clean_temp_folder
from yta_general_utils.programming.path import get_project_abspath
from yta_general_utils.path import create_file_abspath
from yta_general_utils.file.reader import read_json_from_file
from yta_general_utils.logger import print_completed
from typing import Union
from moviepy.editor import concatenate_videoclips


class YoutubeAutonomous:
    __database_handler: DatabaseHandler = None
    """
    Object to interact with the database and get and create projects.

    _This parameter is not manually set by the user._
    """
    __segment_validator: SegmentValidator = None
    """
    Object to validate the segments we want to use to build a project.

    _This parameter is not manually set by the user._
    """
    __segments_abspath: Union[str, None] = None
    """
    The absolute path that points to the segments folder in which this
    software will store the segments related content parts that are
    created during the whole segment building process. This folder 
    allows us recovering a non-finished project in which we generated
    some segments in the past but not all of them.

    _This parameter is not manually set by the user._
    """

    def __init__(self):
        """
        Initializes the object and creates the segment part building
        folder in which segment parts will be stored.
        """
        self.__database_handler = DatabaseHandler()
        self.__segment_validator = SegmentValidator()

        self.__segments_abspath = get_project_abspath() + DEFAULT_SEGMENT_PARTS_FOLDER + '/'
        # We force to create the folder if it doesn't exist
        create_file_abspath(self.__segments_abspath + 'toforce')

    def purge(self, do_remove_segments_files = False):
        """
        Cleans the temporary folder removing all previous generated 
        temporary files, and also the segment files if the
        'do_remove_segments_files' parameter is set to True.
        """
        clean_temp_folder()
        if do_remove_segments_files:
            # TODO: Remove all files in self.__segments_abspath folder
            pass

    def check_config(self):
        # TODO: Check that he config is ok
        pass
     
    def insert_project_in_database_from_file(self, filename: str):
        """
        Reads the provided project content 'filename' and creates a new 
        project in the database if the provided 'filename' contains a new
        project and is valid. If the information belongs to an already
        registered project, it will raise an exception.

        This method returns the new stored project mongo ObjectId if 
        successfully stored, or raises an Exception if anything went wrong.
        """
        if not filename:
            raise Exception('We need a project content "filename" to create a new project.')
        
        if not isinstance(filename, str):
            raise Exception('The "filename" parameter provided is a str.')
        
        json_data = self.__read_project_from_file(filename)

        # If a project with the same content exists, it is the same project
        db_project = self.__database_handler.get_database_project_from_json(json_data)
        if db_project:
            raise Exception('There is an existing project in the database with the same content.')

        db_project = self.__database_handler.insert_project(json_data)

        print_completed('Project created in database with ObjectId = "' + str(db_project['_id']) + '"')

        return str(db_project['_id'])
    
    def process_unfinished_projects(self):
        db_projects = self.get_unfinished_projects()

        for db_project in db_projects:
            # TODO: This should go to a 'Project' class
            print(db_project)
            self.process_project(db_project)

    # TODO: Move this to a new 'Project' class
    def process_project(self, db_project):
        project = Project(db_project['_id'], db_project['status'], db_project['segments'])
        project.build(f'project_{project.id}.mp4')

        # segments = []
        # for index, db_segment in enumerate(db_project['segments']):
        #     segment = Segment(db_segment, str(db_project['_id']), index)
        #     print(segment.text)
        #     segment.build()
        #     # TODO: When finished, change status in the 'segment.py'
        #     segments.append(segment)

        # # TODO: Concatenate all segments in one file
        # full_clips = [segment.full_clip for segment in segments]
        # project_clip = concatenate_videoclips(full_clips)
        # project_clip.write_videofile(f'project_{str(db_project["_id"])}.mp4')
    
    # TODO: This below is for testing, remove it
    def get_unfinished_project(self):
        """
        Returns the first unfinished project that exists in the
        database, or None if there are no unfinished projects.
        """
        return self.__database_handler.get_unfinished_project()

    def get_unfinished_projects(self):
        """
        Returns a list containing all the existing projects in the 
        database that have not been finished yet, or an empty list
        if there are no unfinished projects.
        """
        return self.__database_handler.get_unfinished_projects()
    
    def __read_project_from_file(self, filename: str):
        """
        Reads the provided 'filename', that should contain the information of
        a project, and validates its content and structure. This method will
        return the data read as a dict or will raise an Exception if something
        is not ok.
        """
        if not filename:
            raise Exception('No "filename" provided.')
        
        if not isinstance(filename, str):
            raise Exception('The "filename" parameter provided is a str.')
        
        json_data = read_json_from_file(filename)

        if not 'segments' in json_data:
            raise Exception('The provided "filename" does not contain the expected data structure.')

        # We validate each segment to be able to store the project
        # This will raise an Exception if something is not ok
        for segment in json_data['segments']:
            self.__segment_validator.validate(segment)

        return json_data

    # TODO: Apply ObjectId type to 'project_id'
    def __get_project_from_database_by_id(self, project_id):
        """
        Gets the project with the provided 'project_id' if existing, or
        None if not.
        """
        return self.__database_handler.get_database_project_from_id(project_id)
    
    # TODO: Apply json type to 'json'
    def __get_project_from_database_by_json(self, json):
        """
        Gets the project with the provided 'json' if existing, or None if
        not. This 'json' will be compared with the 'script' field stored
        in database to check if there is a similar project previously 
        stored that must be recovered instead of duplicating it.
        """
        return self.__database_handler.get_database_project_from_json(json)



from youtube_autonomous.shortcodes.enums import ShortcodeType


class Shortcode:
    """
    Class that represent a shortcode detected in a text, containing 
    its attributes and values and also the content if block-scoped.
    """
    def __init__(self, tag: str, type: ShortcodeType, attributes: dict, context, content: str):
        """
        The shortcode has a 'type' and could include some 'attributes' that
        are the parameters inside the brackets, that could also be simple or
        include an associated value. If it is a block-scoped shortcode, it
        will have some 'content' inside of it.
        """
        if not tag:
            raise Exception('No "tag" provided.')
        
        if not type:
            raise Exception('No "type" provided.')
        
        # TODO: Do more checkings

        self.tag = tag
        self.type = type
        # TODO: Maybe do some checks with the attributes, they could be wrong
        # or incomplete. Remember we have 'SegmentElementField' to handle
        # this parameters in a strict way
        self.attributes = attributes
        # TODO: Do we actually need the context (?)
        self.context = context
        # TODO: Do we actually need the content (?)
        self.content = content

        # TODO: We need to include more information to handle them properly
        # 'shortcode': 'sticker',
        # 'previous_word_index': -1,
        self.previous_start_word_index = None
        self.previous_end_word_index = None

        # TODO: Maybe we need to create an abstract Shortcode class to inherit
        # in specific shortcode classes




from youtube_autonomous.exceptions.invalid_segment_exception import InvalidSegmentException
from youtube_autonomous.segments.enums import SegmentType, SegmentField
from youtube_autonomous.segments.validation.consts import VALID_SEGMENT_TYPES, NEED_KEYWORDS_SEGMENT_TYPES, URL_SEGMENT_TYPES, NARRATION_SEGMENT_TYPES, NEED_TEXT_SEGMENT_TYPES
from yta_general_utils.file.checker import file_exists, file_is_audio_file
from yta_general_utils.checker.url import url_is_ok, verify_image_url


class SegmentValidator:
    """
    Class that validates the segment structure and parameters to
    ensure that it is possible to build the content we expect from
    that segment.
    """
    def __init__(self):
        pass

    # TODO: We need to build a 'validate_pre' and a 'validate_post'
    # because we first need to know if we have the basic structure
    # to build a Segment, but then we need to validate that the
    # provided data is also valid (images are downloadable, etc.)

    # TODO: Apply 'segment' type
    def validate(self, segment):
        """
        Validates the provided 'segment' to check if it is a processable one
        so we can store it to be processed later when requested. Checks if 
        the provided 'segment' is valid or not. It returns True if yes or 
        raises a detailed InvalidSegmentException if not.

        The 'segment' parameter must be a dict representing the json that is
        used to define the segment structure and expected content.
        """
        # No type provided
        if not SegmentField.TYPE.value in segment:
            raise InvalidSegmentException('Field "' + SegmentField.TYPE.value + '" not provided.')
        
        # Is a non-accepted type
        if self.__is_invalid_type(segment):
            raise InvalidSegmentException('Type "' + segment[SegmentField.TYPE.value] + '" not valid.')
        
        # Need keywords but doesn't have
        if self.__has_no_keywords(segment):
            raise InvalidSegmentException('No "' + SegmentField.KEYWORDS.value + '" field provided.')

        # Need url but doesn't have
        if self.__has_no_url(segment):
            raise InvalidSegmentException('No "' + SegmentField.URL.value + '" field provided.')
        
        # Need text but doesn't have
        if self.__has_no_text(segment):
            raise InvalidSegmentException('No "' + SegmentField.TEXT.value + '" field provided.')
        
        # Need voice and text at the same time, but only one
        if self.__has_no_voice_and_text(segment):
            raise InvalidSegmentException('Fields "' + SegmentField.VOICE.value + '" and "' + SegmentField.TEXT + '" must exist at the same time to be able to narrate.')
        
        # Need to calculate duration but cannot
        if self.__has_no_duration(segment):
            raise InvalidSegmentException('We need "' + SegmentField.TEXT.value + '", "' + SegmentField.DURATION.value + '" or "' + SegmentField.AUDIO_NARRATION_FILENAME.value + '". We need something to know how much time in some way.')
        
        # Need url but is not available
        if self.__has_invalid_url(segment):
            raise InvalidSegmentException('Provided "' + SegmentField.URL.value + '" is not available.')
        
        # Need image url but is not valid
        if self.__has_no_image_url(segment):
            raise InvalidSegmentException('Provided "' + SegmentField.URL.value + '" is not a valid image.')
        
        # Has narration file but is not valid
        if self.__has_no_valid_file(segment):
            raise InvalidSegmentException('Provided ' + SegmentField.AUDIO_NARRATION_FILENAME.value + ' "' + segment.get(SegmentField.AUDIO_NARRATION_FILENAME.value) + '" is not a valid file.')
        
        # Has narration file but is not audio file
        if self.__has_no_audio_file(segment):
            raise InvalidSegmentException('Provided ' + SegmentField.AUDIO_NARRATION_FILENAME.value + ' "' + segment.get(SegmentField.AUDIO_NARRATION_FILENAME.value) + '" is not an audio file.')
        
        # TODO: Implement content available validation. Check that 'url' 
        # parameter is available and we can obtain the yt video, check
        # that there is a result (a meme, for example) with the given
        # keywords. Check that the youtube video duration is longer than
        # the expected as 'duration' parameter... All those things that
        # are not included above yet

        return True
        
    def __is_invalid_type(self, segment):
        """
        Returns True if the provided 'segment' has an invalid type.
        """
        return segment[SegmentField.TYPE.value] not in VALID_SEGMENT_TYPES

    def __has_no_keywords(self, segment):
        """
        Returns True if the provided 'segment' needs the keywords field but
        doesn't have it.
        """
        return segment[SegmentField.TYPE.value] in NEED_KEYWORDS_SEGMENT_TYPES and not segment.get(SegmentField.KEYWORDS.value)

    def __has_no_url(self, segment):
        """
        Returns True if the provided 'segment' needs the url field but doesn't
        have it.
        """
        return segment[SegmentField.TYPE.value] in URL_SEGMENT_TYPES and not segment.get(SegmentField.URL.value)

    def __has_no_text(self, segment):
        """
        Returns True if the provided 'segment' needs the text field but doesn't
        have it.
        """
        return segment[SegmentField.TYPE.value] in NEED_TEXT_SEGMENT_TYPES and not segment.get(SegmentField.TEXT.value)

    def __has_no_voice_and_text(self, segment):
        """
        Returns True if the provided 'segment' is a narration type segment
        but has only one narration field (voice or text) but not both at
        the same time.
        """
        return segment[SegmentField.TYPE.value] in NARRATION_SEGMENT_TYPES and ((segment.get(SegmentField.TEXT.value) and not segment.get(SegmentField.VOICE.value)) or (segment.get(SegmentField.VOICE.value) and not segment.get(SegmentField.TEXT.value)))

    def __has_no_duration(self, segment):
        """
        Returns True if the provided 'segment' has no field to calculate
        the duration.
        """
        return not segment.get(SegmentField.TEXT.value) and not segment.get(SegmentField.DURATION.value) and not segment.get(SegmentField.AUDIO_NARRATION_FILENAME.value)

    def __has_invalid_url(self, segment):
        """
        Returns True if the provided 'segment' has an invalid url.
        """
        return segment.get(SegmentField.TYPE.value) in URL_SEGMENT_TYPES and not url_is_ok(segment.get(SegmentField.URL.value))

    def __has_no_image_url(self, segment):
        """
        Returns True if the provided 'segment' is an image segment but has
        no valid image url.
        """
        # TODO: Maybe we could handle this with an array like the other ones
        # and not as a simple possible value (image) to do in the same way
        return segment.get(SegmentField.TYPE.value) == SegmentType.IMAGE.value and not verify_image_url(segment.get(SegmentField.URL.value))

    def __has_no_valid_file(self, segment):
        """
        Returns True if the provided 'segment' has a narration file but is 
        not a valid file.
        """
        return segment.get(SegmentField.TYPE.value) in NARRATION_SEGMENT_TYPES and segment.get(SegmentField.AUDIO_NARRATION_FILENAME.value) and not file_exists(segment.get(SegmentField.AUDIO_NARRATION_FILENAME.value))

    def __has_no_audio_file(self, segment):
        """
        Returns True if the provided 'segment' has a narration file but is
        not an audio file.
        """
        return segment.get(SegmentField.TYPE.value) in NARRATION_SEGMENT_TYPES and segment.get(SegmentField.AUDIO_NARRATION_FILENAME.value) and not file_is_audio_file(segment.get(SegmentField.AUDIO_NARRATION_FILENAME.value))
    
    # TODO: Keep working with these below:
    # TODO: Check if 'youtube_video' language exist and is valid
    # TODO: Check if 'youtube_video' duration is valid (not str, not negative nor excesive)
    # TODO: Check if effects provided are valid


from youtube_autonomous.segments.segment import Segment
from youtube_autonomous.segments.enums import ProjectStatus, ProjectField, SegmentStatus
from youtube_autonomous.database.database_handler import DatabaseHandler
from moviepy.editor import concatenate_videoclips
from bson.objectid import ObjectId
from typing import Union


class Project:
    """
    Class that represents a whole video Project, containing different
    segments that are used to build consecutively to end being a whole
    video that is this project video.
    """
    _id: str = None
    """
    The stringified mongo ObjectId that identifies this project in the
    database.
    """
    _status: ProjectStatus = None
    """
    The current project status that allows us to know if this project
    has started or not, or even if it has been finished.
    """
    _segments: list[Segment] = None
    """
    The array that contains this project segments that are used to build
    the whole project video.
    """
    __do_use_setter: bool = True
    """
    Internal variable to know if we should use the custom property setters
    or just assign the value directly.

    _This parameter is not manually set by the user._
    """
    __database_handler: DatabaseHandler = None
    """
    Object to interact with the database and get and create projects.

    _This parameter is not manually set by the user._
    """

    def __init__(self, id: Union[str, ObjectId], status: Union[ProjectStatus, str], segments: list[Segment]):
        self.id = id
        # We avoid updating database 'status' at this point
        self._do_use_setter = False
        self.status = status
        self._do_use_setter = True

        # TODO: Simplify this
        tmp_segments = []
        for index, segment in enumerate(segments):
            tmp_segments.append(Segment(segment, self.id, index))
        self.segments = tmp_segments

        self._database_handler = DatabaseHandler()

    @property
    def unfinished_segments(self) -> list[Segment]:
        """
        Returns all this project segments that has not been built at
        all (they are unfinished).
        """
        return [segment for segment in self.segments if segment.status != SegmentStatus.FINISHED]
    
    @property
    def id(self):
        """
        The stringified mongo ObjectId that identifies this project in the
        database.
        """
        return self._id

    @id.setter
    def id(self, id: Union[ObjectId, str]):
        if not id:
            raise Exception('No "id" provided.')

        if not isinstance(id, (str, ObjectId)):
            raise Exception('The "id" parameter is not a string or an ObjectId.')
        
        if isinstance(id, ObjectId):
            id = str(id)

        self._id = id

    @property
    def status(self):
        """
        The current project status that allows us to know if this project
        has started or not, or even if it has been finished.
        """
        return self._status

    @status.setter
    def status(self, status: Union[ProjectStatus, str] = ProjectStatus.TO_START):
        """
        Updates the 'status' property and also updates it in the database.
        """
        if not status:
            raise Exception('No "status" provided.')
        
        if not isinstance(status, (ProjectStatus, str)):
            raise Exception('The "status" parameter provided is not a ProjectStatus nor a string.')
        
        if isinstance(status, str) and not ProjectStatus.is_valid(status):
            raise Exception('The "status" provided string is not a valid ProjectStatus enum value.')
            
        status = ProjectStatus(status)

        self._status = status.value
        if self.__do_use_setter:
            self.__database_handler.__update_project_field(self.id, ProjectField.STATUS, status.value)

    @property
    def segments(self):
        """
        The array that contains this project segments that are used to build
        the whole project video.
        """
        return self._segments

    @segments.setter
    def segments(self, segments: list[Segment]):
        """
        Updates the 'segments' property with the provided 'segments' parameter.
        This method will check that any of the provided segments are Segment
        objects.
        """
        if not segments:
            raise Exception('No "segments" provided.')
        
        if any(not isinstance(segment, Segment) for segment in segments):
            raise TypeError('Some of the given "segments" is not a Segment.')      

        self._segments = segments

    @property
    def _do_use_setter(self):
        """
        Internal variable to know if we should use the custom property setters
        or just assign the value directly.

        _This parameter is not manually set by the user._
        """
        return self.__do_use_setter
    
    @_do_use_setter.setter
    def _do_use_setter(self, value: bool = True):
        if not isinstance(value, bool):
            raise Exception('The "value" parameter provided is not a bool.')

        self.__do_use_setter = value

    @property
    def _database_handler(self):
        """
        Object to interact with the database and get and create projects.

        _This parameter is not manually set by the user._
        """
        return self.__database_handler
    
    @_database_handler.setter
    def _database_handler(self, database_handler: DatabaseHandler):
        if not database_handler:
            raise Exception('No "database_handler" provided.')
        
        if not isinstance(database_handler, DatabaseHandler):
            raise Exception('The "database_handler" parameter provided is not a DatabaseHandler.')
        
        self.__database_handler = database_handler
        
    def build(self, output_filename: str):
        """
        This method will make that all the segments contained in this
        project are built. It will build the unfinished ones and them
        concatenate them in a final video that is stored locally as
        'output_filename'.
        """
        # I make, by now, 'output_filename' mandatory for this purpose
        if not output_filename:
            raise Exception('No "output_filename" provided.')

        self.status = ProjectStatus.IN_PROGRESS

        for segment in self.unfinished_segments:
            segment.build()

        unfinished_segments_len = len(self.unfinished_segments)
        if unfinished_segments_len > 0:
            raise Exception(f'There are {str(unfinished_segments_len)} segments that have not been completely built (unfinished).')
        
        # I put them together in a whole project clip
        final_clip = concatenate_videoclips([segment.full_filename for segment in self.segments])
        final_clip.write_videofile(output_filename)

        self.status = ProjectStatus.FINISHED



MAX_DURATION_PER_IMAGE = 5
"""
The amount of seconds that an image can be used to build a 
video by applying some effect on it.
"""
MIN_DURATION_PER_IMAGE = 2
"""
The minimum amount of seconds that an image must last when 
treating ai_images segment building to ensure that a new
image is needed. If the value is lower than this one, the
previous image will stay longer to fit this short period of
time.
"""
MAX_DURATION_PER_YOUTUBE_SCENE = 5
"""
The maximum number of seconds that we can extract consecutively 
from a YouTube video (related to avoiding copyright issues).
"""
DEFAULT_SEGMENT_PARTS_FOLDER = 'segment_files'
"""
The folder, in the root of the project in which this library is
executed, to maintain segment building part files in order to be
able to work and to recover them later if something goes wrong.
"""


from enum import Enum


class ImageEngine(Enum):
    """
    The engine that is capable of generating AI images.
    """
    PRODIA = 'prodia'
    FLUX = 'flux'
    DEFAULT = PRODIA

class VoiceEngine(Enum):
    """
    The engine that is capable of generation audio voice narrations.
    """
    GOOGLE = 'google'
    MICROSOFT = 'microsoft'
    DEFAULT = GOOGLE
    # TODO: Add more


from youtube_autonomous.segments.builder.ai.enums import ImageEngine, VoiceEngine
from yta_multimedia.image.generation.ai.prodia import generate_image_with_prodia
from yta_multimedia.experimental.image.generation.ai.flux import generate_dev as generate_image_with_flux
from yta_multimedia.audio.voice.generation.tts.google import narrate as narrate_google
from yta_multimedia.audio.voice.generation.tts.microsoft import narrate as narrate_microsoft
from yta_general_utils.temp import create_temp_filename


def create_ai_image(prompt: str, output_filename: str = None, image_engine: ImageEngine = ImageEngine.DEFAULT):
    """
    Creates an AI image with the provided 'prompt' and stores it locally
    as 'output_filename' (if it is not provided it will generate a 
    temporary file) with the also provided 'image_engine'.

    This method returns the 'output_filename' of the generated image.
    """
    if not prompt:
        raise Exception('No "prompt" provided.')
    
    if not output_filename:
        output_filename = create_temp_filename('ai.png')

    if not image_engine:
        image_engine = ImageEngine.DEFAULT

    if image_engine == ImageEngine.PRODIA:
        generate_image_with_prodia(prompt, output_filename)
    elif image_engine == ImageEngine.FLUX:
        generate_image_with_flux(prompt, output_filename)

    return output_filename

def create_ai_narration(text: str, output_filename: str = None, voice_engine: VoiceEngine = VoiceEngine.DEFAULT):
    """
    Creates an audio narration with the given 'text' and stores it
    locally as the 'output_filename' provided (if not provided, it
    generates a temporary file), generated with the also given
    'voice_engine' voice generation engine.

    This method returns the 'output_filename' of the generated audio
    narration.
    """
    if not text:
        raise Exception('No "text" provided.')
    
    if not output_filename:
        output_filename = create_temp_filename('narration.wav')

    if not voice_engine:
        raise Exception('No "voice_engine" provided.')
    
    if not isinstance(voice_engine, VoiceEngine):
        raise Exception('The "voice_engine" parameter is not a VoiceEngine.')

    # TODO: What about language and voices (?)
    if voice_engine == VoiceEngine.GOOGLE:
        narrate_google(text, output_filename = output_filename)
    elif voice_engine == VoiceEngine.MICROSOFT:
        narrate_microsoft(text, output_filename = output_filename)
    # TODO: Add more

    return output_filename